import fnmatch, functools
import boto3
from typing import Union, Tuple, List, Dict, Union
from aws_lambda_powertools import Logger
from awslambdalawm.security.authz.__impl.domain import (
    PdpRequest, ResourceOperation,
    PdpResponse, 
    Policy, PolicyType, SubjectPolicyRule, ResourcePolicyRule, Rule,
    Effect
)
from awslambdalawm.security.authz.__impl.pdprepo import PdpRepo



class PdpService:

    def __init__(self, pdpRepo:PdpRepo):
        self.__pdpRepo = pdpRepo
        self.__logger = Logger()

    def __evaluateResourceOperation(self, resourceOperation:ResourceOperation, 
            subjectUris:List[str], resourcePolicy:Policy, 
            subjectPolicyRules:List[SubjectPolicyRule],
            conditionsContext:Dict) -> PdpResponse:
        def __ruleActionMatched(rule:Rule) -> bool:
            return functools.reduce(
                lambda acc, ruleActionPattern: acc if acc is True else fnmatch.fnmatch(resourceOperation.action, ruleActionPattern),
                rule.actions,
                False
            )
        def __conditionsMatched(rule:Rule) -> bool:
            def __eval_condition(acc:bool, condition:str): 
                if acc is False: # no point proceeding since multiple condition entries are ANDed
                    return acc
                try:
                    return acc and eval(condition, {"__builtins__": {}}, conditionsContext)
                except Exception as e: # any exception will result in False match
                    self.__logger.warning(f"Error evaluation condition in policy rule '{rule}', error={e}")
                    return False
            return functools.reduce(
                __eval_condition,
                rule.conditions,
                True
            )
        # print(f"{resourcePolicy=}")
        resourcePolicyRules = resourcePolicy.rules if not resourcePolicy is None else []
        # print(f"{resourcePolicyRules=}")
        resourceOperationResult = PdpResponse(
            effect = None, 
            obligations = []
        )
        # print(f"{subjectUris=}")
        # print(f"{resourceOperation=}")
        # print(f"{subjectPolicyRules=}")
        for resourcePolicyRule in resourcePolicyRules:
            if resourceOperationResult.effect != Effect.DENY:
                actionMatched = __ruleActionMatched(resourcePolicyRule)
                subjectUriMatched = functools.reduce(
                    lambda acc, ruleSubjectPattern: acc if acc is True else functools.reduce(
                        lambda acc1, subjectUri: acc1 if acc1 is True else fnmatch.fnmatch(subjectUri, ruleSubjectPattern),
                        subjectUris,
                        False 
                    ),
                    resourcePolicyRule.subjects,
                    False
                )
                conditionsMatched = __conditionsMatched(resourcePolicyRule)
                # print(f"in resource policy rule {actionMatched=}, {subjectUriMatched=}, {resourcePolicyRule=}")
                if actionMatched and subjectUriMatched and conditionsMatched:
                    resourceOperationResult = PdpResponse(
                        effect = resourcePolicyRule.effect,
                        obligations = resourceOperationResult.obligations + resourcePolicyRule.obligations
                    )
        # print(f"after resoource policy rule, {resourceOperationResult=}")
        for subjectPolicyRule in subjectPolicyRules:
            if resourceOperationResult.effect != Effect.DENY:
                actionMatched = __ruleActionMatched(subjectPolicyRule)
                resourceUriMatched = functools.reduce(
                    lambda acc, ruleResourcePattern: acc if acc is True else fnmatch.fnmatch(resourceOperation.resourceUri, ruleResourcePattern),
                    subjectPolicyRule.resources,
                    False
                )
                conditionsMatched = __conditionsMatched(subjectPolicyRule)
                # print(f"in subject policies rules {actionMatched=}, {resourceUriMatched=}, {subjectPolicyRule=}")
                if actionMatched and resourceUriMatched and conditionsMatched:
                    resourceOperationResult = PdpResponse(
                        effect = subjectPolicyRule.effect,
                        obligations = resourceOperationResult.obligations + subjectPolicyRule.obligations
                    )
        return resourceOperationResult

    def decide(self, request:PdpRequest) -> PdpResponse:
        subjectUris = request.subjectUris if not request.subjectUris is None else []
        subjectUris = subjectUris if isinstance(subjectUris, list) else [subjectUris]
        resourceOperations = request.resourceOperations
        resourceOperations = resourceOperations if isinstance(resourceOperations, list) else [resourceOperations]
        resourceUris = [
            resourceOperation.resourceUri 
            for resourceOperation in resourceOperations
        ]
        # get all policies that match subject uris or resource uris specified in the list of resource operations
        relevantPolicies:List[Policy] = self.__pdpRepo.getPolicies(subjectUris, resourceUris)
        # print(f"{relevantPolicies=}") 
        resourcePoliciesMap = dict()
        subjectPolicyRules = []
        # for resource policies, create Map of <policy type>-<entity uri> --> Policy
        # for subject policies, extract combined list of rules
        for policy in relevantPolicies:
            if policy.policyType == PolicyType.RESOURCE and policy.entityUri in resourceUris:
                resourcePoliciesMap[f"{policy.policyType.value}-{policy.entityUri}"] = policy
            elif policy.policyType == PolicyType.SUBJECT and policy.entityUri in subjectUris:
                subjectPolicyRules.extend(policy.rules)
        # print(f"{resourcePoliciesMap=}")
        # print(f"{subjectPolicyRules=}")
        resourceOperationsPdpResponseList = [
            self.__evaluateResourceOperation(
                resourceOperation = resourceOperation, 
                subjectUris = subjectUris, 
                resourcePolicy = resourcePoliciesMap[f"{PolicyType.RESOURCE.value}-{resourceOperation.resourceUri}"] if f"{PolicyType.RESOURCE.value}-{resourceOperation.resourceUri}" in resourcePoliciesMap else None,
                subjectPolicyRules = subjectPolicyRules,
                conditionsContext = request.conditionsContext
            )  
            for resourceOperation in resourceOperations
        ]
        # print(f"{resourceOperationsPdpResponseList=}")
        combinedPdpResponse = functools.reduce(
            lambda combinedPdpResponse, resourceOperationResponse: combinedPdpResponse if combinedPdpResponse.effect is Effect.DENY else PdpResponse(
                effect = resourceOperationResponse.effect if not resourceOperationResponse.effect is None else Effect.DENY,
                obligations = [] if resourceOperationResponse.effect is Effect.DENY else combinedPdpResponse.obligations + resourceOperationResponse.obligations
            ),
            resourceOperationsPdpResponseList,
            PdpResponse(
                effect = None,
                obligations = []
            )
        )
        # print(f"{combinedPdpResponse=}")
        return combinedPdpResponse

    def savePolicy(self, policy:Policy) -> Policy:
        self.__logger.info(f"saving policy: {policy}")
        return self.__pdpRepo.savePolicy(policy)

    def getSubjectPolicy(self, subjectUri:str) -> Policy:
        self.__logger.info(f"retrieving subject policy with uri: {subjectUri}")
        return self.__pdpRepo.getSubjectPolicy(subjectUri)

    def getResourcePolicy(self, resourceUri:str) -> Policy:
        self.__logger.info("retrieving resource policy with uri: {resourceUri}")
        return self.__pdpRepo.getResourcePolicy(resourceUri)


import inspect
from dataclasses import dataclass
from typing import List, Tuple, Union, Callable
from aws_lambda_powertools import Logger
from awslambdalawm.security import (
    Principal,
    context
)
from awslambdalawm.security.authz.__impl.domain import (
    ResourceOperation,
    PdpRequest, 
    Effect
)
from awslambdalawm.security.authz.__impl.exception import AccessDeniedException
from awslambdalawm.security.authz.__impl.pdpservice import PdpService



class Pep:

    def __init__(self, pdpService:PdpService):
        self.pdpService = pdpService
        self.__logger = Logger()

    def pep(
        self,
        resourceOperationsGenerator: Callable[..., Union[ResourceOperation, List[ResourceOperation]]],
        subjectUrisGenerator: Callable[..., List[str]] = None,
        conditionsContextGenerator: Callable[..., dict] = None
    ) -> None:
        roGenFullArgSpec = inspect.getfullargspec(resourceOperationsGenerator)
        roGenParams = roGenFullArgSpec.args
        roGenKwargs = roGenFullArgSpec.varkw
        suGenFullArgSpec = inspect.getfullargspec(subjectUrisGenerator) if not subjectUrisGenerator is None else None
        suGeneratorParams = suGenFullArgSpec.args if not subjectUrisGenerator is None else None
        suGeneratorKwargs = suGenFullArgSpec.varkw if not subjectUrisGenerator is None else None
        ccGenFullArgSpec = inspect.getfullargspec(conditionsContextGenerator) if not conditionsContextGenerator is None else None
        ccGeneratorParams = ccGenFullArgSpec.args if not conditionsContextGenerator is None else None
        ccGeneratorKwargs = ccGenFullArgSpec.varkw if not conditionsContextGenerator is None else None
        def wrapper(fn):
            fullArgSpec = inspect.getfullargspec(fn)
            params = fullArgSpec.args
            varargs = fullArgSpec.varargs
            varkwargs = fullArgSpec.varkw
            defaults = fullArgSpec.defaults if not fullArgSpec.defaults is None else [] 
            defaultsLength = len(defaults)
            paramsLength = len(params)
            defaultsMap = {}
            for i, default in enumerate(defaults):
                defaultsMap[params[paramsLength-defaultsLength+i]] = default
            def innerWrapper(*args, **kwargs):
                targetFnKwArgs = {}
                # if target function has no kwargs, we need to filter args to only those expected by the target function
                if varkwargs is None: 
                    for (k,_) in kwargs.items():
                        if k in params:
                            targetFnKwArgs[k] = kwargs[k]
                else: # else its ok to just take in all the kwargs
                    targetFnKwArgs = kwargs.copy()
                generatorKwargs = targetFnKwArgs.copy() if not targetFnKwArgs is None else {}
                argsLength = len(args)
                # first go through all normal parameters
                for i, param in enumerate(params):
                    if i < argsLength:
                        generatorKwargs[param] = args[i]
                    elif param in targetFnKwArgs:
                        generatorKwargs[param] = targetFnKwArgs[param]
                    elif param in defaultsMap:
                        generatorKwargs[param] = defaultsMap[param]
                # next add varargs
                if varargs and argsLength > paramsLength:
                    generatorKwargs[varargs] = tuple(
                        [
                            args[i]
                            for i in range(paramsLength, argsLength)
                        ]
                    )
                if "principal" not in generatorKwargs:
                    generatorKwargs["principal"] = context.getPrincipal() 
                subjectUris = []
                if not subjectUrisGenerator is None:
                    # subject generator kwargs based on the parameters in the function and principal in context
                    subjectGeneratorKwargs = {}
                    if suGeneratorKwargs is None:
                        for param in suGeneratorParams:
                            subjectGeneratorKwargs[param] = generatorKwargs[param]
                    else:
                        subjectGeneratorKwargs = generatorKwargs.copy()
                    subjectUris = subjectUrisGenerator(**subjectGeneratorKwargs)
                else: 
                    # default based on principal in context
                    currentPrincipal:Principal = context.getPrincipal()
                    if not currentPrincipal is None:
                        subjectUris.append(f"subject://{currentPrincipal.tenant}/{currentPrincipal.subjectType}/{currentPrincipal.subjectId}")
                        if not currentPrincipal is None:
                            roles = currentPrincipal.roles
                            for role in roles:
                                subjectUris.append(f"subject://{currentPrincipal.tenant}/role/{role}")
                # resource generator kwargs based on the parameters in the function and principal in context
                resourceOperationsGeneratorKwargs = {}
                if roGenKwargs is None:
                    for param in roGenParams:
                        resourceOperationsGeneratorKwargs[param] = generatorKwargs[param]
                else:
                    resourceOperationsGeneratorKwargs = generatorKwargs.copy()
                resourceOperations = resourceOperationsGenerator(**resourceOperationsGeneratorKwargs)
                # condition context generator kwargs based on parameters in the function and principal in context
                conditionsContext = {}
                if not conditionsContextGenerator is None:
                    conditionsContextGeneratorKwargs = {}
                    if ccGeneratorKwargs is None:
                        for param in ccGeneratorParams:
                            conditionsContextGeneratorKwargs[param] = generatorKwargs[param]
                    else:
                        conditionsContextGeneratorKwargs = generatorKwargs.copy()
                    conditionsContext = conditionsContextGenerator(**conditionsContextGeneratorKwargs)
                pdpRequest = PdpRequest(
                    subjectUris = subjectUris,
                    resourceOperations = resourceOperations,
                    conditionsContext = conditionsContext
                )
                self.__logger.info(pdpRequest)
                pdpResponse = self.pdpService.decide(request = pdpRequest)
                self.__logger.info(pdpResponse)
                if pdpResponse.effect != Effect.ALLOW:
                    raise AccessDeniedException(f"Access denied for {pdpRequest}")
                else:
                    return fn(*args, **targetFnKwArgs)
            return innerWrapper
        return wrapper




import os
from dataclasses import dataclass, field
from typing import Dict, ClassVar, List
import jwt
import awslambdalawm.security as security

class PrincipalParser:

    def __init__(self, 
            jwtIdTokenSubjectIdField:str,
            jwtIdTokenTenantField:str,
            jwtIdTokenRolesField:str,
            jwtSystemAccessTokenScopePrefix:str
    ):
        self.__jwtIdTokenSubjectIdField = jwtIdTokenSubjectIdField
        self.__jwtIdTokenTenantField = jwtIdTokenTenantField
        self.__jwtIdTokenRolesField = jwtIdTokenRolesField
        self.__jwtSystemAccessTokenScopePrefix = jwtSystemAccessTokenScopePrefix

    def getPrincipalFromEvent(self, event:Dict) -> security.Principal:
        if "headers" not in event or not "authorization" in event["headers"]:
            return None
        authorizationHeader = event["headers"]["authorization"]
        if authorizationHeader is None:
            return None
        jwtToken = authorizationHeader[len("Bearer "):]
        decodedJwt = jwt.decode(jwtToken, options={"verify_signature": False})
        tokenUse = decodedJwt["token_use"]
        if tokenUse == "id":
            clientId = decodedJwt["aud"]
            subjectType = security.Principal.SUBJECT_TYPE_USER
            subjectId = decodedJwt[self.__jwtIdTokenSubjectIdField]
            tenant = decodedJwt[self.__jwtIdTokenTenantField]
            rolesStr = decodedJwt[self.__jwtIdTokenRolesField]
            roles = rolesStr.split(",") if len(rolesStr) > 0 else []
            principal = security.Principal(
                clientId = clientId,
                subjectType = subjectType, 
                subjectId = subjectId, 
                tenant = tenant,
                token = jwtToken,
                roles = roles
            )    
            return principal    
        elif tokenUse == "access":
            subjectId = decodedJwt["sub"]
            subjectType = security.Principal.SUBJECT_TYPE_SYSTEM
            # sample valid scope: https://projectx.platform-dev/projectxb2c
            scope = decodedJwt["scope"]
            tenantPrefixSplit = scope.rsplit("/", 1)
            [prefix, tenant] = tenantPrefixSplit      
            if prefix != self.__jwtSystemAccessTokenScopePrefix:
                raise Exception("Invalid scope in token")
            principal = security.Principal(
                clientId = subjectId,
                subjectType = subjectType, 
                subjectId = subjectId, 
                tenant = tenant,
                token = jwtToken
            )
            return principal
        else:
            raise Exception(f"Invalid token use {tokenUse} encountered")



import os
from dataclasses import dataclass, field
from typing import Dict, ClassVar, List
import jwt

@dataclass
class Principal:
    clientId:str
    subjectType:str
    subjectId:str
    tenant:str
    token:str
    roles:List[str] = field(default_factory=list)
    SUBJECT_TYPE_USER:ClassVar[str] = "user"
    SUBJECT_TYPE_SYSTEM:ClassVar[str] = "system"
    DELEGATED_ACCESS:ClassVar[str] = "delegated-access"


from awslambdalawm.security.core.__impl.domain import Principal

class __Context:
    def __init__(self):
        self.principal = None
    
    def setPrincipal(self, principal:Principal):
        self.principal = principal

    def getPrincipal(self):
        return self.principal

__context = __Context()

def setPrincipal(principal:Principal) -> None:
    __context.setPrincipal(principal)

def getPrincipal() -> Principal:
    return __context.getPrincipal()


import os
import inspect
import json
from datetime import date, datetime
from collections import namedtuple
from typing import Dict, Callable
from dataclasses import asdict, is_dataclass
import jwt
import dacite
from aws_lambda_powertools import Logger
from awslambdalawm.util import DatetimeUtil, ObjectUtil
from awslambdalawm.httpapi.__impl.domain import RequestMethod, Request, Response
from awslambdalawm.httpapi.__impl.principalparser import PrincipalParser
import awslambdalawm.security as security

class App:    
    __FnSpec = namedtuple("FnSpec", ("fn", "principalRequired", "requestBodyType"))  

    def __init__(self, principalParser:PrincipalParser, logger:Logger = None):
        super().__init__()
        self.__functionRegistry = {}
        self.__logger = Logger() if logger is None else logger
        self.__principalParser = principalParser

    def route(self, 
        path:str, 
        method:RequestMethod, 
        principalRequired:bool = True,
        requestBodyType:type = None
    ) -> Callable:
        def wrapper(fn):
            self.__logger.info(f"Adding to function registry: key={method.name} {path}, fn={fn}, principalRequired={principalRequired}, requestBodyType={requestBodyType}")
            self.__functionRegistry[f"{method.name} {path}"] = App.__FnSpec(fn, principalRequired, requestBodyType)     
            return fn
        return wrapper

    def handle(self, event:Dict, context:Dict) -> Dict:
        routeKey = event["routeKey"]
        self.__logger.info(f"Request for routeKey={routeKey} received")
        targetFnSpec:App.__FnSpec = self.__functionRegistry[routeKey]
        targetFn = targetFnSpec.fn
        argspec = inspect.getfullargspec(targetFn)
        targetFnArgNames = argspec[0]
        targetFnKwargs = argspec.varkw
        newKwargs = {}
        newVarargs = []
        if "event" in targetFnArgNames or not targetFnKwargs is None:
            newKwargs["event"] = event
        if "context" in targetFnArgNames or not targetFnKwargs is None:
            newKwargs["context"] = context
        # detect principal and inject if in argument list
        principal = None
        try:
            principal = self.__principalParser.getPrincipalFromEvent(event)
        except Exception as e:
            self.__logger.exception(f"Error parsing principal from request: {e}")
        security.context.setPrincipal(principal)
        if principal is None:
            self.__logger.info(f"No valid principal detected")
            if targetFnSpec.principalRequired:        
                # return error
                return(self.__generateLambdaResponse(Response(401)))
        else:
            self.__logger.info(f"Detected principal: tenant={principal.tenant}, subjectType={principal.subjectType}, subjectId={principal.subjectId}")
        if "principal" in targetFnArgNames or not targetFnKwargs is None:
            self.__logger.info(f"Injecting principal into route handler")
            newKwargs["principal"] = principal   
        if "request" in targetFnArgNames or not targetFnKwargs is None:
            pathParams = event["pathParameters"] if "pathParameters" in event else {}
            queryParams = event["queryStringParameters"] if "queryStringParameters" in event else {}
            requestBody = json.loads(event["body"]) if "body" in event else None
            if requestBody and targetFnSpec.requestBodyType:
                requestBody = ObjectUtil.fromDict(
                    dictValue = requestBody,
                    objectType = targetFnSpec.requestBodyType
                ) 
            request = Request(
                pathParams = pathParams,
                queryParams = queryParams,
                body = requestBody
            )
            self.__logger.info(f"Injecting request with pathParams={pathParams} and queryParams={queryParams} into route handler")
            newKwargs["request"] = request
        self.__logger.info(f"Invoking handler={targetFn} for routeKey={routeKey}")
        try:
            targetFnReturnValue = targetFn(*newVarargs, **newKwargs)
            return self.__generateLambdaResponse(targetFnReturnValue)
        except security.AccessDeniedException:
            return self.__generateLambdaResponse(
                Response(statusCode=403)
            )

    def __generateLambdaResponse(self, response:[object, Response]) -> dict:
        statusCode = 200
        responseBodyObject = response
        if isinstance(response, Response):
            httpResponse:Response = response
            statusCode = httpResponse.statusCode
            responseBodyObject = httpResponse.responseBody
        responseBodyStr = responseBodyObject
        if responseBodyObject and not isinstance(responseBodyObject, str):
            if is_dataclass(responseBodyObject):
                responseBodyStr = ObjectUtil.toJsonStr(asdict(responseBodyObject))
            else:
                responseBodyStr = ObjectUtil.toJsonStr(responseBodyObject)
        return {
            "isBase64Encoded": False,
            "statusCode": statusCode,
            "headers": { 
                "Content-Type": "application/json"
            },
            "body": responseBodyStr
        } 


from datetime import datetime, timezone

def nowUtc() -> datetime:
    return datetime.now(tz=timezone.utc)

def toIso8601Str(dt:datetime, timespec:str="milliseconds") -> str:
    return dt.isoformat(timespec=timespec)

def fromIso8601Str(dtStr:str) -> datetime:
    if dtStr.endswith("Z"):
        dtStr = dtStr[0:len(dtStr)-1] + "+00:00"
    return datetime.fromisoformat(dtStr)


from awslambdalawm.security.core.__impl.domain import Principal

class PermissionsHelper:
    @staticmethod
    def hasPermissionToAny(principal:Principal, *authorisedRoles):
        # system client has full access to its own tenant
        if principal.subjectType == Principal.SUBJECT_TYPE_SYSTEM: 
            return True
        for role in principal.roles:
            if role in authorisedRoles:
                return True
        return False

    @staticmethod
    def hasPermissionToAll(principal:Principal, *requiredRoles):
        # system client has full access to its own tenant
        if principal.subjectType == Principal.SUBJECT_TYPE_SYSTEM: 
            return True
        return (
            len(requiredRoles) == 0 
        ) or (
            len(requiredRoles) > 0 and set(requiredRoles).issubset(set(principal.roles))
        )



from dataclasses import dataclass as __dataclass
from aws_lambda_powertools.logging import Logger as __Logger
from awslambdalawm.httpapi.__impl.app import App as HttpApiApp
from awslambdalawm.httpapi.__impl.principalparser import PrincipalParser as __PrincipalParser
from awslambdalawm.httpapi.__impl.domain import (
    RequestMethod as HttpRequestMethod, 
    Request as HttpRequest,
    Response as HttpResponse
)

@__dataclass
class __Instances:
    httpApiApp:HttpApiApp = None

__Instances = __Instances()

def configure(
    jwtIdTokenSubjectIdField:str,
    jwtIdTokenTenantField:str,
    jwtIdTokenRolesField:str,
    jwtSystemAccessTokenScopePrefix:str,
    logger:__Logger = None
) -> None:
    principalParser = __PrincipalParser(
        jwtIdTokenSubjectIdField=jwtIdTokenSubjectIdField,
        jwtIdTokenTenantField=jwtIdTokenTenantField,
        jwtIdTokenRolesField=jwtIdTokenRolesField,
        jwtSystemAccessTokenScopePrefix=jwtSystemAccessTokenScopePrefix,
    )
    if logger is None:
        __Instances.httpApiApp = HttpApiApp(principalParser=principalParser)
    else:
        __Instances.httpApiApp = HttpApiApp(principalParser=principalParser, logger=logger) 


def httpApiApp() -> HttpApiApp:
    return __Instances.httpApiApp


from enum import Enum
from dataclasses import dataclass
from typing import Dict, Any

class RequestMethod(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    CONNECT = "CONNECT"
    TRACE = "TRACE"

@dataclass
class Request:
    pathParams:Dict[str,str]
    queryParams:Dict[str,str]
    body:Any

@dataclass
class Response:
    statusCode:int
    responseBody:object = None


import json
from datetime import date, datetime
from enum import Enum
from typing import List, Dict
import dacite
import awslambdalawm.util.__impl.datetimeutil as datetimeutil

def fromDict(dictValue:Dict, objectType:type):
    return dacite.from_dict(
        data_class=objectType,
        data = dictValue,
        config = dacite.Config(
            type_hooks = {
                datetime: datetimeutil.fromIso8601Str,
                date: date.fromisoformat
            },
            cast = [
                Enum
            ]
        )        
    )

def toDict(o:object) -> Dict:
    resultDict = {}
    if "__dict__" in dir(o):
        for k,v in o.__dict__.items():
            if (isinstance(v, datetime)):
                resultDict[k] = datetimeutil.toIso8601Str(v)
            elif (isinstance(v, date)):
                resultDict[k] = v.isoformat()
            elif (isinstance(v, Enum)):
                resultDict[k] = v.name
            elif "__dict__" in dir(v):
                resultDict[k] = toDict(v)
            elif isinstance(v, list) or isinstance(v, tuple) or isinstance(v, set):
                resultDict[k] = []
                for v1 in iter(v):
                    resultDict[k].append(toDict(v1))
            else:
                resultDict[k] = v
        return resultDict 
    else:
        return o

def toJsonStr(obj:object) -> str:
    def default_fn(o):
        if isinstance(o, datetime):
            return datetimeutil.toIso8601Str(o)
        if isinstance(o, date):
            return o.isoformat()
        return o.__dict__
    return json.dumps(
        obj,
        default=default_fn
    )


from dataclasses import dataclass as __dataclass
from awslambdalawm.security.core.__impl import PermissionsHelper
import awslambdalawm.security.core.__impl.context as context
from awslambdalawm.security.core.__impl.domain import Principal
from awslambdalawm.security.authz.__impl.domain import ResourceOperation
from awslambdalawm.security.authz.__impl.domain import Policy, Effect, PolicyType, SubjectPolicyRule, ResourcePolicyRule
from awslambdalawm.security.authz.__impl.exception import AccessDeniedException
from awslambdalawm.security.authz.__impl.pep import Pep
from awslambdalawm.security.authz.__impl.pdprepo import PdpRepo as __PdpRepo
from awslambdalawm.security.authz.__impl.pdpservice import PdpService

@__dataclass
class __Instances:
    pdpService:PdpService=None,
    pep:Pep=None

__INSTANCES = __Instances()

def configure(
    policiesTableName:str, 
    policiesTableEntityUriFieldName:str = "entityUri",
    policiesTablePolicyTypeFieldName:str = "policyType",
    policiesKeyPrefix:str = "POLICY#"
) -> None:
    pdpRepo = __PdpRepo(
        policiesTableName,
        policiesTableEntityUriFieldName,
        policiesTablePolicyTypeFieldName,
        policiesKeyPrefix
    )
    __INSTANCES.pdpService = PdpService(pdpRepo)
    __INSTANCES.pep = Pep(__INSTANCES.pdpService)

def pep() -> Pep:
    return __INSTANCES.pep

def pdpService():
    return __INSTANCES.pdpService


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="awslambda-python-common", 
    version="0.0.57",
    author="Lawrence M",
    author_email="lawrence.mok@gmail.com",
    description="AWS Lambda Python development framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/lm-private/awslambda-python-common",
    packages=setuptools.find_packages(
        where="src/main"
    ),
    package_dir={
        "": "src/main/"
    },
    install_requires=[
        "pyjwt",
        "dacite",
        "aws_lambda_powertools"
    ],
    python_requires='>=3.6'
)


from dataclasses import dataclass, field
from enum import Enum
from typing import List, Union, Dict

class Effect(Enum):
    ALLOW = "ALLOW"
    DENY = "DENY"

class PolicyType(Enum):
    SUBJECT = "SUBJECT"
    RESOURCE = "RESOURCE"

@dataclass
class Rule:
    name:str
    effect:Effect
    actions:List[str]

@dataclass
class ResourcePolicyRule(Rule):
    subjects:List[str]
    conditions:List[str] = field(default_factory=list)
    obligations:List[str] = field(default_factory=list)

@dataclass
class SubjectPolicyRule(Rule):
    resources:List[str]
    conditions:List[str] = field(default_factory=list)
    obligations:List[str] = field(default_factory=list)

@dataclass
class Policy:
    name:str
    entityUri:str
    policyType:PolicyType
    rules:List[Union[SubjectPolicyRule,ResourcePolicyRule]]

@dataclass(unsafe_hash=True)
class ResourceOperation:
    action:str
    resourceUri:str

@dataclass
class PdpRequest:
    subjectUris:List[str]
    resourceOperations:List[ResourceOperation]
    conditionsContext:Dict = field(default_factory=dict)

@dataclass
class PdpResponse:
    effect:Effect
    obligations:List[str]


from typing import List, Dict
import boto3
from boto3.dynamodb.conditions import Key
from aws_lambda_powertools.logging import Logger
from awslambdalawm.util import ObjectUtil
from awslambdalawm.security.authz.__impl.domain import Policy, SubjectPolicyRule, ResourcePolicyRule, PolicyType, Effect

class PdpRepo:

    __DYNAMODB_RESOURCE = boto3.resource("dynamodb")

    def __init__(self, 
            policiesTableName:str, 
            policiesTableEntityUriFieldName:str = "entityUri",
            policiesTablePolicyTypeFieldName:str = "policyType",
            policiesKeyPrefix:str = ""):
        self.__policiesTableName = policiesTableName
        self.__policiesTableEntityUriFieldName = policiesTableEntityUriFieldName
        self.__policiesTablePolicyTypeFieldName = policiesTablePolicyTypeFieldName
        self.__policiesKeyPrefix = policiesKeyPrefix 
        self.__policiesDdbTableResource = PdpRepo.__DYNAMODB_RESOURCE.Table(policiesTableName)
        self.__logger = Logger()

    def savePolicy(self, policy:Policy) -> None:
        policyDict = ObjectUtil.toDict(policy)
        policyDict[self.__policiesTableEntityUriFieldName] = self.__prefixKeyFieldValue(policy.entityUri)
        policyDict[self.__policiesTablePolicyTypeFieldName] = self.__prefixKeyFieldValue(policy.policyType)
        self.__policiesDdbTableResource.put_item(
            Item = policyDict
        )
        self.__logger.info(f"savePolicy: saved policy {policyDict}")
        return policy

    def getPolicies(self, subjectUris:List[str], resourceUris:List[str]) -> List[Policy]:
        policiesDict = PdpRepo.__DYNAMODB_RESOURCE.batch_get_item(
            RequestItems = {
                self.__policiesTableName: {
                    "Keys": [
                        {
                            self.__policiesTableEntityUriFieldName: prefixedUri,
                            self.__policiesTablePolicyTypeFieldName: self.__prefixKeyFieldValue(PolicyType.SUBJECT.value)
                        } for prefixedUri in self.__prefixEntityUris(subjectUris)
                    ] + [
                        {
                            self.__policiesTableEntityUriFieldName: prefixedUri,
                            self.__policiesTablePolicyTypeFieldName: self.__prefixKeyFieldValue(PolicyType.RESOURCE.value)
                        } for prefixedUri in self.__prefixEntityUris(resourceUris)
                    ]
                }
            }
        )
        self.__logger.info(f"{policiesDict=}")
        returnPolicies = []
        if self.__policiesTableName in policiesDict["Responses"]:
            policyDicts = policiesDict["Responses"][self.__policiesTableName]
            returnPolicies = self.__policyDictsToPolicies(policyDicts)
        self.__logger.info(f"getPolicies: returning policies {returnPolicies}")
        return returnPolicies

    def getSubjectPolicy(self, subjectUri:str) -> Policy:
        return self.__getPolicy(subjectUri, PolicyType.SUBJECT)

    def getSubjectPolicies(self, subjectUris:List[str]) -> List[Policy]:
        return self.__getPolicies(subjectUris, PolicyType.SUBJECT)

    def getResourcePolicy(self, resourceUri:str) -> Policy:
        return self.__getPolicy(resourceUri, PolicyType.RESOURCE)

    def getResourcePolicies(self, resourceUris:List[str]) -> List[Policy]:
        return self.__getPolicies(resourceUris, PolicyType.RESOURCE)

    def __getPolicies(self, entityUris:List[str], policyType:PolicyType) -> List[Policy]:
        policiesDict = PdpRepo.__DYNAMODB_RESOURCE.batch_get_item(
            RequestItems = {
                self.__policiesTableName: {
                    "Keys": [
                        {
                            self.__policiesTableEntityUriFieldName: prefixedUri,
                            self.__policiesTablePolicyTypeFieldName: self.__prefixKeyFieldValue(policyType.value)
                        } for prefixedUri in self.__prefixEntityUris(entityUris)
                    ]
                }
            }
        )
        returnPolicies=[]
        if self.__policiesTableName in policiesDict["Responses"]:
            policyDicts = policiesDict["Responses"][self.__policiesTableName]
            returnPolicies = self.__policyDictsToPolicies(policyDicts)
        self.__logger.info(f"__getPolicies: returning policies {returnPolicies}")
        return returnPolicies

    def __getPolicy(self, entityUri:str, policyType:PolicyType) -> Policy:
        policyDict = self.__policiesDdbTableResource.get_item(
            Key = {
                self.__policiesTableEntityUriFieldName: self.__prefixKeyFieldValue(entityUri),
                self.__policiesTablePolicyTypeFieldName: self.__prefixKeyFieldValue(policyType.value)
            }
        )
        returnPolicy = None
        if "Item" in policyDict:
            returnPolicy = self.__policyDictToPolicy(policyDict["Item"])
        return returnPolicy

    def __policyDictsToPolicies(self, policyDicts:List[Dict]) -> List[Policy]:
        return [
            self.__policyDictToPolicy(policyDict)
            for policyDict in policyDicts
        ] 

    def __policyDictToPolicy(self, policyDict:Dict) -> Policy:
        return ObjectUtil.fromDict(
            dictValue = policyDict,
            objectType = Policy
        )

    def __prefixEntityUris(self, entityUris:List[str]) -> List[str]:
        return [
            self.__prefixKeyFieldValue(entityUri)
            for entityUri in entityUris
        ]

    def __prefixKeyFieldValue(self, keyFieldValue:str) -> str:
        return f"{self.__policiesKeyPrefix}{keyFieldValue}"

    def __unprefixDdbKeyFieldValue(self, ddbKeyFieldValue:str) -> str:
        return ddbKeyFieldValue.lstrip(self.__policiesKeyPrefix)

# mySubjectPolicy = Policy(
#     entityUri = "subject://projectx/custodian/111-111",
#     policyType = PolicyType.SUBJECT,
#     rules = [
#         SubjectPolicyRule(
#             effect = Effect.ALLOW,
#             resources = [
#                 "resource://projectx/organisation/*"
#             ],
#             actions = [
#                 "read"
#             ]
#         )
#     ]
# )
# savePolicy(mySubjectPolicy)

# retrievedSubjectPolicy = getSubjectPolicy("subject://projectx/custodian/111-111")
# print(f"{retrievedSubjectPolicy=}")

# myResourcePolicy = Policy(
#     entityUri = "resource://projectx/organisation/888",
#     policyType = PolicyType.RESOURCE,
#     rules = [
#         ResourcePolicyRule(
#             effect = Effect.ALLOW,
#             subjects = [
#                 "subject://projectx/custodian/*",
#                 "subject://projectx/administrator/*"
#             ],
#             actions = [
#                 "*"
#             ]
#         ),
#         ResourcePolicyRule(
#             effect = Effect.ALLOW,
#             subjects = [
#                 "subject://projectx/audience/222-222"
#             ],
#             actions = [
#                 "read"
#             ]
#         )
#     ]
# )
# savePolicy(myResourcePolicy)

# retrievedResourcePolicy = getResourcePolicy("resource://projectx/organisation/888")
# print(f"{retrievedResourcePolicy=}")

# print(getSubjectPolicies(["subject://projectx/custodian/111-111"]))